/* (C) 13.0.3.10.12 Thoth
 *
 * Contributions by:
 *    Mario Ricardo Rodriguez
 *    Richard Andrews
 *    Luca Deri
 *    Lukasz Taczuk
 */

#if !defined( UON_WIRE_H_ )
#define UON_WIRE_H_

#include <stdlib.h>

#if defined(WIN32)
#include "win32/uon_win32.h"

#if defined(__MINGW32__)
#include <stdint.h>
#endif /* #ifdef __MINGW32__ */

#else /* #if defined(WIN32) */
#include <stdint.h>
#include <netinet/in.h>
#include <sys/socket.h> /* AF_INET and AF_INET6 */
#endif /* #if defined(WIN32) */

#define UON_PKT_VERSION                 2
#define UON_DEFAULT_TTL                 2       /* can be forwarded twice at most */
#define UON_COMMUNITY_SIZE              16
#define UON_MAC_SIZE                    6
#define UON_COOKIE_SIZE                 4
#define UON_PKT_BUF_SIZE                2048
#define UON_SOCKBUF_SIZE                64      /* string representation of INET or INET6 sockets */

typedef uint8_t uon_community_t[UON_COMMUNITY_SIZE];
typedef uint8_t uon_mac_t[UON_MAC_SIZE];
typedef uint8_t uon_cookie_t[UON_COOKIE_SIZE];

typedef char    uon_sock_str_t[UON_SOCKBUF_SIZE];       /* tracing string buffer */

enum uon_pc {
    uon_ping=0,                 /* Not used */
    uon_register=1,             /* Register node to node */
    uon_deregister=2,           /* Deregister this node */
    uon_packet=3,               /* PACKET data content */
    uon_register_ack=4,         /* ACK of a registration from node to node */
    uon_register_super=5,       /* Register node to supernode */
    uon_register_super_ack=6,   /* ACK from supernode to node */
    uon_register_super_nak=7,   /* NAK from supernode to node - registration refused */
    uon_federation=8            /* Not used by node */
};

typedef enum uon_pc uon_pc_t;

#define UON_FLAGS_OPTIONS               0x0080
#define UON_FLAGS_SOCKET                0x0040
#define UON_FLAGS_FROM_SUPERNODE        0x0020

/* The bits in flag that are the packet type */
#define UON_FLAGS_TYPE_MASK             0x001f  /* 0 - 31 */
#define UON_FLAGS_BITS_MASK             0xffe0

#define IPV4_SIZE                       4
#define IPV6_SIZE                       16


#define UON_AUTH_TOKEN_SIZE             32      /* bytes */


#define UON_EUNKNOWN                    -1
#define UON_ENOTIMPL                    -2
#define UON_EINVAL                      -3
#define UON_ENOSPACE                    -4


typedef uint16_t uon_flags_t;
typedef uint16_t uon_transform_t;       /* Encryption, compression type. */
typedef uint32_t uon_sa_t;              /* security association number */

struct uon_sock {
    uint8_t     family;         /* AF_INET or AF_INET6; or 0 if invalid */
    uint16_t    port;           /* host order */
    union 
    {
    uint8_t     v6[IPV6_SIZE];  /* byte sequence */
    uint8_t     v4[IPV4_SIZE];  /* byte sequence */
    } addr;
};

typedef struct uon_sock uon_sock_t;

struct uon_auth {
    uint16_t    scheme;                         /* What kind of auth */
    uint16_t    toksize;                        /* Size of auth token */
    uint8_t     token[UON_AUTH_TOKEN_SIZE];     /* Auth data interpreted based on scheme */
};

typedef struct uon_auth uon_auth_t;


struct uon_common {
    /* int                 version; */
    uint8_t             ttl;
    uon_pc_t            pc;
    uon_flags_t         flags;
    uon_community_t     community;
};

typedef struct uon_common uon_common_t;

struct uon_REGISTER {
    uon_cookie_t        cookie;         /* Link REGISTER and REGISTER_ACK */
    uon_mac_t           srcMac;         /* MAC of registering party */
    uon_mac_t           dstMac;         /* MAC of target node */
    uon_sock_t          sock;           /* REVISIT: unused? */
};

typedef struct uon_REGISTER uon_REGISTER_t;

struct uon_REGISTER_ACK {
    uon_cookie_t        cookie;         /* Return cookie from REGISTER */
    uon_mac_t           srcMac;         /* MAC of acknowledging party (supernode or node) */
    uon_mac_t           dstMac;         /* Reflected MAC of registering node from REGISTER */
    uon_sock_t          sock;           /* Supernode's view of node socket (IP Addr, port) */
};

typedef struct uon_REGISTER_ACK uon_REGISTER_ACK_t;

struct uon_PACKET {
    uon_mac_t           srcMac;
    uon_mac_t           dstMac;
    uon_sock_t          sock;
    uon_transform_t     transform;
};

typedef struct uon_PACKET uon_PACKET_t;


/* Linked with uon_register_super in uon_pc_t. Only from node to supernode. */
struct uon_REGISTER_SUPER {
    uon_cookie_t        cookie;         /* Link REGISTER_SUPER and REGISTER_SUPER_ACK */
    uon_mac_t           nodeMac;        /* MAC to register with node sending socket */
    uon_auth_t          auth;           /* Authentication scheme and tokens */
};

typedef struct uon_REGISTER_SUPER uon_REGISTER_SUPER_t;


/* Linked with uon_register_super_ack in uon_pc_t. Only from supernode to node. */
struct uon_REGISTER_SUPER_ACK {
    uon_cookie_t        cookie;         /* Return cookie from REGISTER_SUPER */
    uon_mac_t           nodeMac;        /* MAC registered to node sending socket */
    uint16_t            lifetime;       /* How long the registration will live */
    uon_sock_t          sock;           /* Sending sockets associated with nodeMac */

    /* The packet format provides additional supernode definitions here. 
     * uint8_t count, then for each count there is one
     * uon_sock_t.
     */
    uint8_t             num_sn;         /* Number of supernodes that were send
                                         * even if we cannot store them all. If
                                         * non-zero then sn_bak is valid. */
    uon_sock_t          sn_bak;         /* Socket of the first backup supernode */

};

typedef struct uon_REGISTER_SUPER_ACK uon_REGISTER_SUPER_ACK_t;


/* Linked with uon_register_super_ack in uon_pc_t. Only from supernode to node. */
struct uon_REGISTER_SUPER_NAK {
    uon_cookie_t        cookie;         /* Return cookie from REGISTER_SUPER */
};

typedef struct uon_REGISTER_SUPER_NAK uon_REGISTER_SUPER_NAK_t;



struct uon_buf {
    uint8_t *   data;
    size_t      size;
};

typedef struct uon_buf uon_buf_t;

int encode_uint8( uint8_t * base, size_t * idx, const uint8_t v );
int decode_uint8( uint8_t * out, const uint8_t * base, size_t * rem, size_t * idx );
int encode_uint16( uint8_t * base, size_t * idx, const uint16_t v );
int decode_uint16( uint16_t * out, const uint8_t * base, size_t * rem, size_t * idx );
int encode_uint32( uint8_t * base, size_t * idx, const uint32_t v );
int decode_uint32( uint32_t * out, const uint8_t * base, size_t * rem, size_t * idx );
int encode_buf( uint8_t * base, size_t * idx, const void * p, size_t s);
int decode_buf( uint8_t * out, size_t bufsize, const uint8_t * base, size_t * rem, size_t * idx );
int encode_mac( uint8_t * base, size_t * idx, const uon_mac_t m );
int decode_mac( uint8_t * out, const uint8_t * base, size_t * rem, size_t * idx ); /* of size UON_MAC_SIZE. This clearer than passing a uon_mac_t */
int encode_common( uint8_t * base, size_t * idx, const uon_common_t * common );
int decode_common( uon_common_t * out, const uint8_t * base, size_t * rem, size_t * idx );
int encode_sock( uint8_t * base, size_t * idx, const uon_sock_t * sock );
int decode_sock( uon_sock_t * sock, const uint8_t * base, size_t * rem, size_t * idx );
int encode_REGISTER( uint8_t * base, size_t * idx, const uon_common_t * common, const uon_REGISTER_t * reg );
int decode_REGISTER( uon_REGISTER_t * pkt, const uon_common_t * cmn, const uint8_t * base, size_t * rem, size_t * idx ); /* info on how to interpret it */
int encode_REGISTER_SUPER( uint8_t * base, size_t * idx, const uon_common_t * common, const uon_REGISTER_SUPER_t * reg );
int decode_REGISTER_SUPER( uon_REGISTER_SUPER_t * pkt, const uon_common_t * cmn, const uint8_t * base, size_t * rem, size_t * idx );
int encode_REGISTER_ACK( uint8_t * base, size_t * idx, const uon_common_t * common, const uon_REGISTER_ACK_t * reg );
int decode_REGISTER_ACK( uon_REGISTER_ACK_t * pkt, const uon_common_t * cmn, const uint8_t * base, size_t * rem, size_t * idx );
int encode_REGISTER_SUPER_ACK( uint8_t * base, size_t * idx, const uon_common_t * cmn, const uon_REGISTER_SUPER_ACK_t * reg );
int decode_REGISTER_SUPER_ACK( uon_REGISTER_SUPER_ACK_t * reg, const uon_common_t * cmn, const uint8_t * base, size_t * rem, size_t * idx );
int fill_sockaddr( struct sockaddr * addr, size_t addrlen, const uon_sock_t * sock );
int encode_PACKET( uint8_t * base, size_t * idx, const uon_common_t * common, const uon_PACKET_t * pkt );
int decode_PACKET( uon_PACKET_t * pkt, const uon_common_t * cmn, const uint8_t * base, size_t * rem, size_t * idx ); /* info on how to interpret it */


#endif /* #if !defined( UON_WIRE_H_ ) */
